#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pthread.h>
#include <time.h>
#include <semaphore.h>


#include <errno.h>


#define Travel_ag_num 5// καθορισμός των travel agent
#define SEATS 100// καθορισμός των θέσεων



// δημιουργία δομής 
typedef struct 
{
    sem_t mutex;
    int books;// για την αποθήκευση του αριθμού των κρατήσεων
    int bookings[SEATS];//για την ενημέρωση έαν μία θέση είναι πιασμένη
    int booking_time[SEATS];//για την αποθήκευση του χρόνου κράτησης
}Booking;



int main(){

    // δημιουργία αρχείου για την αποθήκευση των τιμών με σκοπό την δημιουργία γραφήματος 
    FILE *fp = fopen("booking_data.dat","w");
    if (fp==NULL)// έλεγχος για τυχών λάθη
    {
        perror("failed to open file");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    
    


    int pipefds[Travel_ag_num][2];// πίνακας των pipes για την ενημέρωση των παιδιών
    int shmid;
    Booking *booking;
    time_t start_time, current_time; // για να πάρουμε το χρόνο κάθε κράτησης
    
    

    struct timespec start, finish; 
    clock_gettime(CLOCK_REALTIME, &start);// Για μέτρηση χρόνου εκτέλεσης κώδικα

     

    // δημιουργούμαι την κοινή μνήμη και εντοπισμός λαθών
    
    
    shmid = shmget(IPC_PRIVATE, sizeof(Booking), IPC_CREAT | 0666);
    if (shmid == -1)
    {
        perror("shmget");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    // Έπειτα συνδέουμαι την κοινή μνήμη με τις θέσεις όπου έχει γίνει book και εντοπισμός λαθών
    
    booking = (Booking *)shmat(shmid, NULL, 0);
    if (booking == (void *) -1)
    {
        perror("shmat");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }

    //Αρχικοποίηση των θέσεων και εντοπισμός λαθών
    booking->books = 0;
    for (int i = 0; i < SEATS; i++)
    {
        booking->bookings[i] = 0; //όπου 0 ελεύθερη θέση
    }
    if (sem_init(&booking->mutex, 1, 1)!=0)
    {
        perror("sem_init");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    

    printf("Initialized bookings. All seats are free.\n");
    fflush(stdout);
    

    // Δημιουργία των 1D pipes για την ενημέρωση των παιδιών
    for (int i = 0; i < Travel_ag_num; i++)
    {
        if (pipe(pipefds[i]) == -1)//εντοπισμός λαθών
        {
            perror("pipe");
            fflush(stdout);
            exit(EXIT_FAILURE);
        }
        
        
    }


     srand(time(NULL));

    for (int i = 0; i < Travel_ag_num; i++)
    {
       
       
        pid_t pid = fork();

        if (pid <0)
        { //έλεγχος για εντοπισμός λαθών
            perror("fork");
            fflush(stdout);
            exit(1);
        }
        

        else if (pid == 0)
        { //child proccess
        
            close(pipefds[i][1]); // κλείσιμο εγραφών 
            int seat; 
            start_time = time(NULL);
            
            // έλεγχος δεδομένων απο το pipe για το έαν πληρούνται οι προυποθέσεις 
            while (read(pipefds[i][0], &seat, sizeof(seat))> 0)
            {
                
                
                printf("Attempting to acquire semaphore\n");
                fflush(stdout);
                sem_wait(&booking->mutex);// κλείδωμα με semaphore
                printf("semaphore acquired \n");
                fflush(stdout);


                if (booking->books >= SEATS)// έξτρα έλεγχος για το έαν υπάρχει θέση
                {
                   
                    printf("releasing semaphore 2\n");
                    fflush(stdout);
                    sem_post(&booking->mutex);
                    printf("semaphore releasing 2 \n");
                    fflush(stdout);
                    break;
                }
                
            
               
                
                if (booking->bookings[seat] == 0)// ελεγχός εάν η θέση είναι ελεύθερη
                {
                    booking->bookings[seat] = 1; // κράτηση θέσης 
                    booking->books++;// άυξηση του μετρητή θέσεων όπου έχει γίνει book
                    current_time = time(NULL);
                    booking->booking_time[seat] = (int)(current_time-start_time);//Αποθήκευση χρόνου που χρειάστηκε για τη κράτηση
                    printf("Agent %d booked seat %d. Remaining: %din time %d\n", i+1, seat+1, SEATS - booking->books, booking->booking_time[seat]);
                    fflush(stdout);
                    // for (int k = 0; k < SEATS; k++) // για debag που χρειάστηκε
                    // {
                    //     book_time[k]=booking->booking_time[k];
                    //     printf("%d\n",booking->booking_time[k]);
                    // }
                    //δημιουργία notify-send 
                    char notification[256];
                    sprintf(notification, "notify-send 'Booking Update' 'Seat %d booked at %d sec'", seat+1, booking->booking_time[seat]);
                    system(notification);
                    // έλεγχος για το έαν έχει γίνει book σε όλες τις θέσεις για την κατάλληλη εκτύπωση
                    if (booking->books == SEATS)  
                    {
                        printf("Flight is full. \n");
                        break;
                    }
                    
                }

                printf("releasing semaphore 1\n");
                fflush(stdout);
                sem_post(&booking->mutex);// ξεκλέιδωμα της semaphore
                printf("semaphore releasing 1 \n");
                fflush(stdout);
                
                // πρόβλεψη του χρόνου εκτέλεσης του προγράμματος
                int avg_time = (int)((current_time-start_time)/booking->books);
                int predicted_time = avg_time* SEATS;
                printf("predict_time %d\n", predicted_time);
                fflush(stdout);
                //δημιουργία notify-send 
                char prediction[256];
                sprintf(prediction, "notify-send 'Prediction' 'All seats will booked in approx %d sec'", predicted_time);
                system(prediction);
                
                    
             
             
            }

            close(pipefds[i][0]);// κλείσιμο αχρησιμοποίητων διαβασμάτων 
            exit(0);
            
        }
        else if (pid > 0)
        {
            printf("close remain\n");
            fflush(stdout);
            close(pipefds[i][0]); // κλείσιμο αχρησιμοποίητων διαβασμάτων 
            
        }
        
    }
    

    
   
   
    for (int j = 0; j < SEATS; j++)
    {
        
        int agent = rand() % Travel_ag_num;
        write(pipefds[agent][1], &j, sizeof(j));//τυχαίοι πράκτορες για να κάνουν τη κράτηση
        usleep(rand()% 3000000 + 1 );// τυχαία καθυστέρηση 
        
        
    }

    for (int i = 0; i < Travel_ag_num; i++)
    {//parent proccess
        close(pipefds[i][1]);
        close(pipefds[i][0]);
        
    }

    //περιμένουμε τα παιδία να τερματήσουν
    for (int i = 0; i < Travel_ag_num; i++)
    {
        siginfo_t info;
        waitid(P_PID, getegid(), &info, WNOHANG);
    }
    

       
    // εγγραφή δεδομένων στο αρχείο και έλεγχος για λάθη
   for (int k = 0; k < SEATS; k++)
    {   // αριθμός κρατήσεων και χρόνος
        if(fprintf(fp,"%d %d\n",k+1, booking->booking_time[k])<0){
            perror("Failed to write file");
            fflush(stdout);
            fclose(fp);// κλείσιμο αρχείου
            exit(EXIT_FAILURE);
                
            }
        }
    fclose(fp);// κλείσιμο αρχείου 
    
    

    //Αποσύνδεση και αφαίρεση της κοινής μνήμης και εντοπισμός λαθών
    sem_destroy(&booking->mutex);
    if (shmdt(booking) == -1)
    {
        perror("shmdt");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    
    if (shmctl(shmid, IPC_RMID, NULL) == -1)
    {
        perror("shmctl");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    printf("Shared memory segment removed successfully.\n");
    fflush(stdout);

    // παίρνουμε το χρόνο μέχρι την εκτέλεση των proccess
    clock_gettime(CLOCK_REALTIME, &finish);
    long seconds = finish.tv_sec - start.tv_sec; 
    long ns = finish.tv_nsec - start.tv_nsec; 

     if (start.tv_nsec > finish.tv_nsec) { // clock underflow 
	--seconds; 
	ns += 1000000000; 
    } 
    printf("seconds without ns: %ld\n", seconds); 
    fflush(stdout);
    printf("nanoseconds: %ld\n", ns); 
    fflush(stdout);
    printf("total seconds: %e\n", (double)seconds + (double)ns/(double)1000000000); 
    fflush(stdout);


    // Δημιουργία γραφίματος και παραμέτροποίηση αυτού όπως ζητήθηκε και έλεγχος για λάθη
    int status =system("xmgrace -settype xy -block booking_data.dat -graph 0 -bxy 1:2 -pexec 'xaxis label \"Time (s)\"; yaxis label \"Amplitude (units)\"' -autoscale xy -saveall plot.arg");
    if (status == -1)
    {
        perror("Failed to execute system command");
        fflush(stdout);
        exit(EXIT_FAILURE);
    }

    return 0;
    
}